/*********************************************************************/
/*                                                                   */
/*   Program Name :  data_recov                                      */
/*                                                                   */
/*      Version:  1.00  2006-07-24                                   */
/*                                                                   */
/*                                                                   */
/*      Header data edit or recovering for K5/VSSP data              */
/*                                                                   */
/*                                                                   */
/*      Copyright (c) 2006 T.Kondo/NICT All Right Reserved           */
/*                                                                   */
/*********************************************************************/
/*
      usage

         data_recov K5file sfreq adbit numch start_time
       
              where  K5file   ----- K5 data file name to edit
                     sfreq    ----- Sampling frequency (MHz)
                                    0.04,0.1,0.2,0.5,1,2,4,8,16,32,64
                     adbit    ----- A/D resolution (1,2,4,8)
                     numch    ----- # of channels (1,4)
                     start_time --- 1st time of data
	                                either "HH:MM:SS" or "seconds in day"

        ver 0.00 2006-07-23  start development 
        ver 1.00 2006-07-24  release 


F:/datacheck/t2044_kb1_144-0453_2006144045324.k5
F:/datacheck/t2044_kb1_144-1527_2006144152738.k5
F:/datacheck/t2044_kb1_144-1705a_2006144170520.k5
D:\IPVLBI\data\temp\O202151500a.dat
*/
#define PROGRAM_NAME "data_recov"
#define PROGRAM_VERSION "Ver. 2006-07-24"

#include "libk5.h"      // common header for IP-VLBI utilities

#include "libvssp32.h"

void vssp_hd_make_ini(vssp_head_t *vph)
{
    /* VSSP headder common information set up */
    int n,fid,aid;
    long int w1;
    unsigned char b,c;
    //double fsampl;
    /* sync pattern 1 set */
    for (n=0; n<4; n++){
       vph->bhead[n]=0xFF;
    }
    /* seconds set */
    w1=vph->seconds;
    b=(unsigned char)((w1 & 0x10000) >> 16);
    vph->bhead[4]=(unsigned char)(w1 & 0xFF);
    vph->bhead[5]=(unsigned char)((w1 & 0xFF00) >> 8);
    b=(unsigned char)((w1 & 0x10000) >> 16);
    /* ch flag set */
    if( vph->numch == 4){
        b=b | 0x2;
    }
    /* sampling freq index set */
    fid=0;
    //if( ipb->sfreq ==    40) fid=0;
    if( vph->sfreq ==   100) fid=1;
    if( vph->sfreq ==   200) fid=2;
    if( vph->sfreq ==   500) fid=3;
    if( vph->sfreq ==  1000) fid=4;
    if( vph->sfreq ==  2000) fid=5;
    if( vph->sfreq ==  4000) fid=6;
    if( vph->sfreq ==  8000) fid=7;
    if( vph->sfreq == 16000) fid=8;
    if( vph->sfreq == 32000) fid=9;
    if( vph->sfreq == 64000) fid=10;
    if( vph->sfreq == 128000) fid=11; 
    if( vph->sfreq == 256000) fid=12;
    if( vph->sfreq == 512000) fid=13;
    if( vph->sfreq == 1024000) fid=14;
    if( vph->sfreq == 2048000) fid=15;
    c=(unsigned char)(fid << 2);
    b=b | c;
    /* A/D bit length set */
    aid=0;
    //if( ipb->adbit == 1 ) aid=0;
    if( vph->adbit == 2 ) aid=1;
    if( vph->adbit == 4 ) aid=2;
    if( vph->adbit == 8 ) aid=3;
    c=(unsigned char)(aid << 6);
    b=b | c;
    vph->bhead[6]=b;
    /* sync pattern 2 set */
    //vph->bhead[7]=0x8B;
    vph->bhead[7]=vph->psync2;
    /* internal data initialize */
//    fsampl=1000.0*(double)ipb->sfreq; // sampling frequency (Hz)
//
//    ipb->nbsec=(((long)fsampl)*ipb->adbit*ipb->numch)/8;  // # of bytes in 1 sec data except header
//    ipb->nbsec+=8;
//    ipb->bpos=0;
}

void vssp_hd_make(vssp_head_t *vph)
{
    /* VSSP data headder second data up date */
    long int w1;
    unsigned char b,c;
    /* seconds set */
    w1=vph->seconds;
    b=(unsigned char)((w1 & 0x10000) >> 16);
    vph->bhead[4]=(unsigned char)(w1 & 0xFF);
    vph->bhead[5]=(unsigned char)((w1 & 0xFF00) >> 8);
    b=vph->bhead[6];
    c=(unsigned char)((w1 & 0x10000) >> 16);
    b=(b & 0xFE) | (c & 0x1);
    vph->bhead[6]=b;
}

int cont_file_open(FILE *fi, char *fname, int *f_c, char *k5name, int64_t *size)
{     
       /* continuous file open */
       char fname2[80];
       char cbuf[80];

       *f_c++;
       // continuous file name creation     XXXX.datN   where N=1,2,3......
       strcpy(fname2,fname);
       sprintf(cbuf,"%d\0",*f_c);
       strcat(fname2,cbuf);
       strcpy(k5name,fname2);
       *size=get_file_size(k5name);
       // file open 
       if( (fi  = fopen( k5name, "rb" )) == NULL ) {
           //printf("File Open Error! \n");
           return -1;
       }
       printf("\n continuous file (%s (%d bytes)) opened!\n",k5name,*size);  
       printf("                                    "); 
       return 0;
}



int find_offset(char* fname, vssp_head_t *vph, int mode) 
{
       /* find out possible byte offset */
       int i;
       int64_t tword;
       long int numb,npos, nkai;
       long int bitsec;
       int headerbits;
       int total_frame, bad_frame;
       char k5name[80],cbuf[10];
       double fsample;
       int bytesec,wordsec;
	   u_int32_t ubuf1,ubuf2;
       int wordread;
	   u_int32_t pattern1,pattern2,mask1,mask2;
	   int sa,ipos;
	   int byteoffset,iok;
	   int64_t pos,posold;   // current pointer
	   int64_t size;  // opened file size in bytes **K5**
       FILE *fi;
       //fpos_t size;

       total_frame=0;
       bad_frame=0;
       numb=0;
       npos=0;
       nkai=0;
	   byteoffset=-1;
	   iok=0;

       headerbits=64;  // for K5/VSSP
       if(mode==32) headerbits=96+20*8;  // for K5/VSSP32 usually 256

       /* calculation of datasize in 1sec */

       fsample=1000.0*vph->sfreq;   // sampling frequency (Hz)
       //K5.tsample=1.0/K5.fsample;            // sampling period (sec)


       bitsec=(long)fsample*vph->adbit*vph->numch;  // # of bits in 1 sec data except header
       bytesec=bitsec/8;                  //# of bytes in 1 sec data
       wordsec=bitsec/32;                 //# of bytes in 1 sec data

       strcpy(k5name,fname);

       /* K5/VSSP file open */

       size=get_file_size(k5name);
       if( (fi  = fopen( k5name, "rb" )) == NULL ) {
           printf("File %s Open Error! \n",k5name);
           goto exitend2;
       }
       printf("FILE : %s  (%lld bytes)\n",k5name,size);

       //K5.twordread=0;  // reset word read counter
       pos=0;        // reset current pointer position
	   posold=0;
       tword=0;
       

       /* initial pattern serach (this may be good for k51a only?) */
	   /*
             Broken headder usually has a pattern as follows
			 
			   FFFF???? 8B????FF

                (this may be good for k51a only?)

	   */
	   pattern1=0xFFFF0000;
	   pattern2=0x8B0000FF;
	   mask1 = 0xFFFF0000;
	   mask2 = 0xFF0000FF;

	   printf("Pattern %08X %08X \n",pattern1,pattern2);
	   printf("Mask    %08X %08X searching....\n",mask1,mask2);
	   nkai=0;
       wordread = fread( &ubuf1, sizeof( u_int32_t ), 1, fi );
       while(1){
          wordread = fread( &ubuf2, sizeof( u_int32_t ), 1, fi );
          if( ferror( fi ) || wordread==0 ){ // File EOF or read error
             goto exitend;
          }
          if(((ubuf1 & mask1)== (pattern1 & mask1))
			  &&((ubuf2 & mask2)== (pattern2 & mask2))){
			  iok=1;
		      sa=(int)(pos-posold);
		      cbuf[0]='\0';
              if(sa == bytesec+headerbits/8) strcpy(cbuf,"GOOD");
			  ipos=(int)pos;
              printf("%5d Pattern found %08X %08X at pos=%12d  sa=%10d %s\n",
				  nkai,ubuf1,ubuf2,ipos,sa,cbuf);
              //printf("Posold=%lld\n",posold );
              //printf("sa=%d\n",sa );
			  if(nkai == 0) byteoffset=ipos;
		      posold=pos;
			  nkai++;
			  if(nkai==10) goto exitend;
          }
          pos+=4;
		  ubuf1=ubuf2;
       }
exitend:
       i=fclose( fi );
	   if(iok==1){
		   printf(" Found byte offset is %d\n",byteoffset);
	   } else {
		   printf(" Sorry! No pattern found! I gave up!\n");
	   }
exitend2:
	   return byteoffset;

}


void checkK5file(char* fname, vssp_head_t *vph, int mode, int byteoffset) 
{
       int i;
       int64_t tword;
       long int bitsec;
       int headerbits;
       int f_c;
       int total_frame;
       char k5name[80];
       double fsample;
       int bytesec,wordsec;
       u_int32_t ubuf;
       int wordread;
	   int64_t pos,posold;   // current pointer
	   int64_t size;  // opened file size in bytes **K5**
       FILE *fi;
       //fpos_t size;

       total_frame=0;
       f_c=0;

       headerbits=64;  // for K5/VSSP
       if(mode==32) headerbits=96+20*8;  // for K5/VSSP32 usually 256

       /* calculation of datasize in 1sec */

       fsample=1000.0*vph->sfreq;   // sampling frequency (Hz)
       //K5.tsample=1.0/K5.fsample;            // sampling period (sec)


       bitsec=(long)fsample*vph->adbit*vph->numch;  // # of bits in 1 sec data except header
       bytesec=bitsec/8;                  //# of bytes in 1 sec data
       wordsec=bitsec/32;                 //# of bytes in 1 sec data


       strcpy(k5name,fname);

       /* K5/VSSP file open */

       size=get_file_size(k5name);
       if( (fi  = fopen( k5name, "rb" )) == NULL ) {
           printf("File %s Open Error! \n",k5name);
           goto exitend2;
       }
       printf("FILE : %s  (%lld bytes)\n",k5name,size);

       //K5.twordread=0;  // reset word read counter
       pos=0;        // reset current pointer position
	   posold=0;
       tword=0;
       
       /* move pointer to initial position */
       if(byteoffset > 0){
          fseek(fi,byteoffset,SEEK_CUR); // move pointer by K5->bytesec
          if( ferror( fi ) ){
              printf("set initial position error!\n");
              exit(0);
          }
       }
       pos+=byteoffset;   // current pointer position update



loop1:

       /* supposed header data read */

       /* 64bit or 256bit data read by 32bit step */

       printf("%04d    ",total_frame);
       for(i=0; i<headerbits/32; i++){
          wordread = fread( &ubuf, sizeof( u_int32_t ), 1, fi );
          if( ferror( fi ) || wordread==0 ){ // File EOF or read error
             i=fclose( fi ) ;
             // next file candidate open
             i=cont_file_open(fi,fname,&f_c,k5name,&size);
             if(i < 0) goto exitend2;
             pos=0;
             wordread = fread( &ubuf, sizeof( u_int32_t ), 1, fi );
             if( ferror( fi ) || wordread==0 ) goto exitend;
          }
          printf(" %08X",ubuf); 
          pos+=4;  // 4 bytes count up
       }
       printf("\n");


       /* move 1sec data */

       pos+=bytesec;  // pointer position to be set
       if(pos < size){  // not yet reached to EOF
          fseek(fi,bytesec,SEEK_CUR); // move pointer by bytesec
          if( ferror( fi ) ){
             i=fclose( fi ) ;
             goto exitend2;
          }
       } else {  // exceed EOF so pointer set to EOF
          //fseek(fi,(int)(size - pos),SEEK_CUR); // 
          //if( ferror( fi ) ){
          //   i=fclose( fi ) ;
          //   goto exitend2;
          //}
          pos=pos-size;  // position to be set in the next file
          i=fclose( fi ) ;
          // next file candidate open
          i=cont_file_open(fi,fname,&f_c,k5name,&size);
          if(i < 0) goto exitend2;
          // point next header
          fseek(fi,(int)pos,SEEK_CUR); // 
          if( ferror( fi ) ){
             i=fclose( fi ) ;
             goto exitend2;
          }
       }

       total_frame++;

       goto loop1;
   
 exitend:
       i=fclose( fi ) ;
 exitend2:;
       printf("\n");
}



void editK5file(char* fname, char* oname, vssp_head_t *vph, int mode, int byteoffset) 
{
       //int usize=20000;  // ubuf size
       int usize;  // ubuf size
       int i,n,k;
       int64_t tword;
       long int bitsec;
       int headerbits;
       int f_c;
       int total_frame, bad_frame;
       char k5name[80];
       double fsample;
       int bytesec,wordsec,imax;
       u_int32_t *ubuf;
	   u_int32_t *head;  
       int wordread,iwrite;
	   int64_t pos,posold;   // current pointer
	   int64_t size;  // opened file size in bytes **K5**
       FILE *fi,*fo;

       total_frame=0;
       bad_frame=0;

       f_c=0;

       headerbits=64;  // for K5/VSSP
       if(mode==32) headerbits=96+20*8;  // for K5/VSSP32 usually 256

       /* calculation of datasize in 1sec */

       fsample=1000.0*vph->sfreq;   // sampling frequency (Hz)
       //K5.tsample=1.0/K5.fsample;            // sampling period (sec)


       bitsec=(long)fsample*vph->adbit*vph->numch;  // # of bits in 1 sec data except header
       bytesec=bitsec/8;                  //# of bytes in 1 sec data
       wordsec=bitsec/32;                 //# of bytes in 1 sec data

       // buffer size fix so as wordsec/usize to become integer number 
       usize=1250;
       if(vph->sfreq >= 100) usize=3125;
       if(vph->sfreq >= 1000) usize=31250;
       imax=wordsec/usize;

       ubuf=(u_int32_t *)malloc(usize*sizeof(u_int32_t));

       strcpy(k5name,fname);

       /* K5/VSSP file open */

       size=get_file_size(k5name);
       if( (fi  = fopen( k5name, "rb" )) == NULL ) {
           printf("File %s Open Error! \n",k5name);
           goto exitend2;
       }
       printf("FILE : %s  (%lld bytes)\n",k5name,size);

	   /* Output file open */
       if( (fo  = fopen( oname, "wb" )) == NULL ) {
           printf("File %s Open Error! \n",oname);
           goto exitend2;
       }
	   /* Header initialize */
       vssp_hd_make_ini(vph);

       //K5.twordread=0;  // reset word read counter
       pos=0;        // reset current pointer position
	   posold=0;
       tword=0;
       
       /* move pointer to initial position */
       if(byteoffset > 0){
          fseek(fi,byteoffset,SEEK_CUR); // move pointer by K5->bytesec
          if( ferror( fi ) ){
              printf("set initial position error!\n");
              exit(0);
          }
       }
       pos+=byteoffset;   // current pointer position update



loop1:

       /* supposed header data read */

       /* 64bit or 256bit data read by 32bit step */

       printf("%04d    ",total_frame);
       for(i=0; i<headerbits/32; i++){
          wordread = fread( ubuf, sizeof( u_int32_t ), 1, fi );
          if( ferror( fi ) || wordread==0 ){ // File EOF or read error
             i=fclose( fi ) ;
             // next file candidate open
             i=cont_file_open(fi,fname,&f_c,k5name,&size);
             if(i < 0) goto exitend2;
             pos=0;
             wordread = fread( ubuf, sizeof( u_int32_t ), 1, fi );
             if( ferror( fi ) || wordread==0 ) goto exitend;
          }
          printf(" %08X",ubuf[0]); 
          pos+=4;  // 4 bytes count up
       }

       vssp_hd_make(vph);
	   // header write 
       iwrite=fwrite(vph->bhead,sizeof(char),headerbits/8,fo);
	   printf(" ===> ");
	   head=(u_int32_t *)vph->bhead;
       for(i=0; i<headerbits/32; i++){
          printf(" %08X",head[i]);
	   }		  
       printf("\n");


	   vph->seconds++;   // seconds update

       /* copy 1sec data */

	   for(n=0; n<imax; n++){
          wordread = fread( ubuf, sizeof( u_int32_t ), usize, fi );
          if( ferror( fi ) || wordread==0 ){ // File EOF or read error
			 if(wordread > 0) 
             i=fclose( fi ) ;
             // next file candidate open
             i=cont_file_open(fi,fname,&f_c,k5name,&size);
             if(i < 0) goto exitend2;
			 // rest of data read
             k = fread( ubuf, sizeof( u_int32_t ), usize-wordread, fi );
             iwrite=fwrite(ubuf,sizeof(u_int32_t),k,fo);
		  } else {
             iwrite=fwrite(ubuf,sizeof(u_int32_t),wordread,fo);
		  }
	   }

       total_frame++;

       goto loop1;
   
 exitend:
       i=fclose( fi ) ;
 exitend2:;
       i=fclose( fo ) ;
       free(ubuf);
       printf("\n");
}

int hhmmss2sec(char *cbuf, int *hh, int *mm, int *ss)
{
	//  HH:MM:SS => seconds
	int n,len;
	int hh1,mm1,ss1,seconds;
	int p1,p2, n1st;
	char buf[10];
	len=strlen(cbuf);
    // check ':' position
	p1=0; p2=0;
	n1st=0;
    //printf("len %d\n",len);
	for(n=0; n<len; n++){
		if(cbuf[n]==':'){
			if(n1st==0){
				p1=n;
				n1st=1;
			} else {
				p2=n;
				n1st=2;
			}
		}
	}
	//printf("p1,p2 %d %d\n",p1,p2);
	if(n1st==0){ // no HH:MM:SS
       seconds=atoi(cbuf);
	   hh1=seconds/3600;
	   mm1= (seconds % 3600)/60;
	   ss1= seconds % 60;
       *hh=hh1; *mm=mm1; *ss=ss1;
	   return seconds;
	}

	strncpy(buf,cbuf,p1);
	buf[p1]='\0';
	hh1=atoi(buf);
	strncpy(buf,cbuf+p1+1,p2-p1-1);
	buf[p2-p1]='\0';
	mm1=atoi(buf);
	strncpy(buf,cbuf+p2+1,len-p2-1);
	buf[len-p2]='\0';
	ss1=atoi(buf);
    seconds=hh1*3600+mm1*60+ss1;
	*hh=hh1; *mm=mm1; *ss=ss1;
	//printf("HH MM SS %d %d %d\n",hh,mm,ss); // debug
	return seconds;
}




         

int main(int argc, char *argv[])
{
    char fname[80];
	char oname[100];
	char cbuf[20];
	char cw;
    int mode, byteoffset;
	int hh,mm,ss;
	double fsmhz;
    double telapsed;           // to measure processing time
    vssp_head_t vph;
    clock_t cstart1sec, cfinish1sec;  // to calculate processing time
    mode=0;
    //if(argc < 2) {
    if(argc < 6) {
       //strcpy(fname,dflt_fname);
       printf("%s %s\n",PROGRAM_NAME,PROGRAM_VERSION);
       printf("How to run: \n");
       printf("  data_recov K5file sfreq adbit numch start_time\n");
       printf("\n");
       printf("    where  K5file ----- K5 data file name to edit\n");
       printf("           sfreq    ----- Sampling frequency (MHz)\n");
       printf("                          0.04,0.1,0.2,0.5,1,2,4,8,16,32,64\n");
       printf("           adbit    ----- A/D resolution (1,2,4,8)\n");
       printf("           numch    ----- # of channels (1,4)\n");
       printf("           start_time --- 1st time of data\n");
	   printf("                           either ""HH:MM:SS"" or ""seconds in day""\n");
       //printf("           mode      ---  32: K5/VSSP32,  0: K5/VSSP (default)\n");
       printf("\n");
       printf("    created filename will be datafile+"".edit""\n");
       exit(0);
    }
    if (argc>=2){
       strcpy(fname,*(argv+1));
    } 
    if(argc >=3){
       fsmhz = atof(*(argv + 2 ));
	   vph.sfreq=(int)(fsmhz*1000.0);
    }
    if(argc >=4){
       vph.adbit = atoi(*(argv + 3 ));
    }
    if(argc >=5){
       vph.numch = atoi(*(argv + 4 ));
    }
    if(argc >=6){
       strcpy(cbuf,*(argv + 5 ));
       vph.seconds=hhmmss2sec(cbuf,&hh,&mm,&ss);
       //printf("seconds=%d\n",vph.seconds); // debug
       //printf("hh mm ss %d %d %d\n",hh,mm,ss); // debug

    }
	// out file name creation
	strcpy(oname,fname);
	strcat(oname,".edit");
	//printf(" Out file name: %s\n",oname);
	// set 2nd sync pattern
	vph.psync2=0x8B;              // K5/VSSP
	if(mode==32) vph.psync2=0x8C; // K5/VSSP32
    byteoffset=0;  // initial read offset in byte
	//byteoffset=865584;

    printf("************ %s  %s ************\n",PROGRAM_NAME,PROGRAM_VERSION);
	printf("\n");
	printf(" Run condition is as follows\n");
	printf("   Input  K5 file           : %s\n",fname);
	printf("   K5 file after edit       : %s\n",oname);
	printf("   Sampling Frequency (MHz) : %f\n",fsmhz);
	printf("   AD bits                  : %d\n",vph.adbit);
	printf("   # of channels            : %d\n",vph.numch);
	printf("   Start Time               : %02d:%02d:%02d (%d seconds)\n",hh,mm,ss,vph.seconds);
	printf("\n");
	printf(" Is this condition OK? ([Y]/N) ");
    cw=getchar();  // read to accept just return
    if(cw=='N' || cw=='n') return 0;

    // processing start
	cstart1sec = clock();
    byteoffset=find_offset(fname,&vph,mode); 
	if(byteoffset < 0) return 0;
	printf("\n Continue Processing with offset %d? ([Y]/N) ",byteoffset);
    cw=getchar();  // read to accept just return
    if(cw=='N' || cw=='n') return 0;
	printf("\n");
	printf("  1 -- Header Check Only using %d byte offset\n",byteoffset);
	printf("  2 -- Replace Header    using %d byte offset\n",byteoffset);
	printf("\n Enter your selection ([1]/2) ");
    cw=getchar();  // read to accept just return
    if(cw=='2'){
       editK5file(fname,oname,&vph,mode,byteoffset);
	} else {
       checkK5file(fname,&vph,mode,byteoffset);
	}
    cfinish1sec = clock();
    telapsed = (double)(cfinish1sec - cstart1sec) / CLOCKS_PER_SEC;
    printf("Time elapsed for processing is %f sec\n",telapsed);
    return 0;
}
